import logging
import uuid

from .entities import LcEntity
from .anchors import Anchor

from antelope import comp_dir, check_direction, PropertyExists, RxRef


class InvalidParentChild(Exception):
    pass


class BalanceAlreadySet(Exception):
    pass


class CacheAlreadySet(Exception):
    pass


class ScenarioConflict(Exception):
    pass


class Fragment(LcEntity):
    """
    What can I say? THey're complicated.
    Conceptually, a fragment is a modeler's tool for keeping track of measurements of a specific material flow.

    A fragment's main purpose is to express a linkage between a parent node and a target or 'anchor' node by means
    of an observed flow.

    Fragments have the following broad functionalities:

     - they can be named. external_ref can be set. not clear why this is only for fragments, but it is meant to
     provide h-r semantic consistency / clarity during data use
     - they can be observed. a fragment has a cached and an observed exchange value. the cached value is meant to
     indicate its value upon creation, and observed is the default value returned upon quantitative queries.
     observations take the form of a scenario specification and an exchange value (along with metadata such as dqi,
     uncertainty, etc)
     - they can be anchored, to data sets or to other models.  anchor points are also specified by scenario. anchors
     can be toggled to "descend" (i.e. expand sub-fragments) or non-descend / roll-up sub-fragments
     - they have child flows. "reference flows" have no parent and supply/consume a good or service through exchange
     with a second party.  Child flows are then dependent on the reference flow. By traversing these links to their
     ends, the entire "span" of the study can be enumerated. It is the job of the traversal engine to ensure that
     traversal terminates upon reaching a background node (i.e. one with membership in a cycle) to a flat LCI.
     - they can balance. Each fragment node (the thing that is 'anchored') can compute a balance during traversal,
     assigning a designated child flow the magnitude of the balance at traversal time.  The quantity that is conserved
     is precisely the designated balance flow's reference quantity.  all flows are queried for characterization
     in computing the balance.
    * STILL TBD: how to handle "balancing" into and out of anchored subfragments
    """
    @classmethod
    def new(cls, flow, direction, **kwargs):
        """
        :param flow: entity or ref
        :param direction: "Input" or "Output" w/r/t parent (future: "balance" to be also accepted)
        :param kwargs: parent; exchange_value/observe; anchor/anchor_flow/descend; balance_flow (alt.), external_ref
        :return:
        """
        return cls(uuid.uuid4(), flow, direction, **kwargs)

    _ref_field = 'parent'

    _parent = None
    _flow = None
    _direction = None

    _new_fields = ['StageName']

    def __init__(self, the_uuid, flow, direction, parent=None, exchange_value=None, units=None, stage_name=None,
                 anchor=None, anchor_flow=None, descend=True,
                 balance_flow=False, external_ref=None, observe=True, **kwargs):

        self._flow = flow
        self._direction = check_direction(direction)
        self._evs = {1: 0.0}
        self._anchors = dict()
        self._child_flows = list()

        super(Fragment, self).__init__('fragment', external_ref, entity_uuid=the_uuid, **kwargs)
        if self._external_ref == self._uuid:
            self._external_ref = None  # reset this
        self['StageName'] = stage_name or ''

        if self.uuid is None:
            self.uuid = str(uuid.uuid4())

        if parent:
            self.set_parent(parent)

        self._balance = False
        self._balance_child = None

        if bool(balance_flow):
            if parent is None:
                logging.warning('Ignoring balance spec for reference flow')
            else:
                self.set_balance_flow()

        if self._balance:
            self._cached_ev = None
            if observe:
                logging.warning('Ignoring observe specification for balance flow')
        else:
            value = exchange_value or 1.0
            if units is not None and len(units) > 0:
                value *= self.conversion(units)

            self._cached_ev = value
            if observe:
                self._evs[1] = self._cached_ev

        if anchor:
            self._anchors[None] = Anchor(self, anchor, anchor_flow=anchor_flow, descend=descend)
        else:
            self._anchors[None] = Anchor.null(self)

    def clear_evs(self, observed=False):
        self._cached_ev = 1.0
        self._evs = {1: int(bool(observed))}

    def __hash__(self):
        """
        Fragments must use UUID as hash because the link is mutable
        :return:
        """
        if self._origin is None:
            raise AttributeError('Origin not set!')
        return hash('%s/%s' % (self.origin, self.uuid))

    @property
    def cached_ev(self):
        return self._cached_ev

    @cached_ev.setter
    def cached_ev(self, value):
        if self.cached_ev != 1.0:
            raise CacheAlreadySet('Set Value: %g (new: %g)' % (self.cached_ev, value))
        if value == self.cached_ev:
            return
        self._cached_ev = value

    @property
    def observed_ev(self):
        return self._evs[1]

    @property
    def external_ref(self):
        if self._external_ref is None:
            return self._uuid
        return self._external_ref

    @external_ref.setter
    def external_ref(self, ref):
        """
        Specify how the entity is referred to in the source dataset. If this is unset, the UUID is assumed
        to be used externally.
        :param ref:
        :return:
        """
        if self._external_ref is None:
            if ref != self.uuid:  # don't bother setting if it's the same as the UUID
                self._external_ref = ref
        else:
            raise PropertyExists('External Ref already set to %s' % self._external_ref)

    def de_name(self):
        """
        Remove a fragment's name
        :return:
        """
        self._external_ref = None

    def reference(self, flow=None):
        """
        For process interoperability
        :return:
        """
        rx = RxRef(self, self.flow, comp_dir(self.direction), self.get('Comment', None), value=self.observed_ev)
        if flow is not None:
            if not rx.flow.match(flow):
                raise ValueError('%.5s: Supplied flow %s does not match fragment' % (self.uuid, flow))
        return rx

    def make_ref(self, query):
        """
        We do NOT want to make local fragments into refs-- but we DO want to make remote fragments into refs. so
        we simply neuter the workhorse function here.
        :param query:
        :return:
        """
        return self

    @property
    def flow(self):
        return self._flow

    @property
    def direction(self):
        return self._direction

    @property
    def anch(self):
        return self._anchors[None]

    '''
    Parents and Child Flows
    '''
    def top(self):
        if self.reference_entity is None:
            return self
        return self.reference_entity.top()

    def set_parent(self, parent):
        if self.reference_entity is not None:
            self.unset_parent()
        if self.origin != parent.origin:
            if self.origin is None:
                self.origin = parent.origin
            else:
                raise AttributeError('Origin mismatch: parent (%s) vs child (%s)' % (parent.origin, self.origin))
        self._set_reference(parent)
        parent.add_child(self)

    def unset_parent(self):
        if self.is_balance:
            self.unset_balance_flow()
        self.reference_entity.remove_child(self)
        self._set_reference(None)

    def add_child(self, child):
        """
        This should only be called from the child's set_parent function
        :param child:
        :return:
        """
        # if child.reference_entity is not self:
        #     raise InvalidParentChild('Fragment should list parent as reference entity')
        if child not in self._child_flows:
            self._child_flows.append(child)
        if self.anch.is_null:  # formerly to_foreground()
            self.observe_anchor(None, self)
        for term in self._anchors.values():
            term.clear_score_cache()

    def remove_child(self, child):
        """
        This should only be called from the child's unset_parent function
        :param child:
        :return:
        """
        if child.reference_entity is not self:
            raise InvalidParentChild('Fragment is not a child')
        self._child_flows.remove(child)
        if len(self._child_flows) == 0:
            if self.anch.node is self:
                self.clear_anchor()
        for term in self._anchors.values():
            term.clear_score_cache()

    @property
    def child_flows(self):
        for k in self._child_flows:  # sorted(self._child_flows, key=lambda x: x.uuid):
            yield k

    def children_with_flow(self, flow, direction=None, anchor=None, recurse=False):
        for k in self._child_flows:
            if k.flow == flow:
                if direction is not None:
                    if k.direction != direction:
                        continue
                if anchor is not None:
                    if k.anch.node != anchor:
                        continue
                yield k
            if recurse:  # depth-first
                for z in k.children_with_flow(flow, direction, recurse=recurse):
                    yield z

    @property
    def parent(self):
        return self.reference_entity

    @property
    def is_reference(self):
        return self.reference_entity is None

    @property
    def is_background(self):
        return len(self._child_flows) == 0

    @property
    def name(self):
        if self._external_ref is None:
            return self['Name']
        return self.external_ref

    @property
    def dirn(self):
        return {
            'Input': '-<-',
            'Output': '=>='
        }[self.direction]

    @property
    def unit(self):
        """
        used for formatting the fragment in display
        :return:
        """
        if self.reference_entity is None:
            return '%4g %s' % (self.cached_ev, self.flow.unit)
        return self.anch.unit

    def __str__(self):
        if self.reference_entity is None:
            if len(self._child_flows) == 0:
                re = '(B) ref'
            else:
                re = ' ** ref'
        else:
            re = self.reference_entity.uuid[:7]
        if self.external_ref == self.uuid:
            extname = ''
        else:
            extname = '{%s}' % self.external_ref

        return '(%s) %s %.5s %s %s  [%s] %s %s' % (re, self.dirn, self.uuid, self.dirn, self.anch,
                                                   self.unit, self['Name'], extname)

    def show(self):
        print('%s' % self)
        super(Fragment, self).show()
        evs = list(self._evs.keys())
        evs.remove(1)
        print('Exchange values: ')
        print('%20.20s: %g' % ('Cached', self.cached_ev))
        print('%20.20s: %g' % ('Observed', self.observed_ev))
        for k in evs:
            print('%20.20s: %g' % (k, self._evs[k]))
        if self.is_balance:
            print('\nBalance flow: True (%s)' % self.flow.reference_entity)
        else:
            print('\nBalance flow: False')
        print('Terminations: ')
        print('%20s  %s' % ('Scenario', 'Termination'))
        for k, v in self._anchors.items():
            if v.node is self:
                print('%20.20s: %s Foreground' % (k, v))
            else:
                if v.descend:
                    desc = '     '
                else:
                    desc = '(agg)'
                print('%20.20s: %s %s %s' % (k, v, desc, v.node))

    def scale_evs(self, factor):
        """
        needed when foregrounding terminations
        :param factor:
        :return:
        """
        for k, v in self._evs.items():
            self._evs[k] = v * factor
        self._cached_ev *= factor

    def observable(self, scenario=None):
        return self._check_observability(scenario=scenario)

    def _check_observability(self, scenario=None):
        if self.reference_entity is None:
            return True
        elif self.is_balance:
            # print('observability: value set by balance.')
            return False
        elif self.reference_entity.anchor(scenario).is_subfrag:
            # print('observability: value set during traversal')
            return False
        else:
            return True

    def conversion(self, units):
        try:
            return self.flow.reference_entity.convert(units)
        except KeyError:
            logging.warning('Flow conversion error: %5.5s: %s (%s)' % (self.uuid, self.flow.reference_entity,
                                                                       units))
            return 0.0

    def observe(self, scenario=None, value=None, units=None):
        if scenario is None:
            scenario = 1
        if isinstance(scenario, tuple) or isinstance(scenario, set):
            raise ScenarioConflict('Set EV must specify single scenario')
        if value is None:
            self._evs[scenario] = self._cached_ev
        else:
            value = float(value)

            if units is not None and len(units) > 0:
                value *= self.conversion(units)
            self._evs[scenario] = value

    def observe_anchor(self, scenario, anchor_node, anchor_flow=None, descend=None):
        if scenario in ('cached', 'observed'):
            raise ValueError('scenario cannot use reserved name: %s' % scenario)

        if isinstance(scenario, tuple) or isinstance(scenario, set):
            raise ScenarioConflict('Set termination must specify single scenario')
        if scenario is not None and scenario in self._anchors:
            if not self._anchors[scenario].is_null:
                raise CacheAlreadySet('Scenario termination already set. use clear_termination()')

        anchor = Anchor(self, anchor_node, anchor_flow, descend)

        if scenario is None:
            """
            use default anchor to set stagename
            """
            if self['StageName'] == '' and not anchor.is_null:
                if anchor.is_frag:
                    self['StageName'] = anchor.node['StageName']
                elif anchor.is_context:
                    self['StageName'] = anchor.node.name
                else:
                    try:
                        self['StageName'] = anchor.node['Name']
                    except (KeyError, TypeError, IndexError):
                        print('%.5s StageName failed %s' % (self.uuid, anchor.node))
                        self['StageName'] = anchor.node.name

        self._anchors[scenario] = anchor

        return anchor

    def observe_lcia_score(self, scenario, quantity, score):
        try:
            anch = self._anchors[scenario]
        except KeyError:
            anch = self.observe_anchor(scenario, self)
        anch.add_lcia_score(quantity, score, scenario=scenario)

    def _match_scenario_ev(self, scenario):
        if scenario is None:
            return None
        _match = []
        for sc in scenario:
            try:
                m = self._evs[sc]
            except KeyError:
                continue
            _match.append(m)
        if len(_match) > 1:
            raise ScenarioConflict('Fragment %s: Multiple scenario results: %s' % (self.uuid, _match))
        m = _match[0]
        if str(m).startswith('norm') and self.parent is None:
            print('Applying and removing one-time normalization scenario %s' % m)
            # self.dbg_print('Applying and removing one-time normalization scenario %s' % m, level=0)
            scenario.remove(m)
            return m
        if scenario in self._evs:
            return scenario
        return 1

    def _match_scenario_anchor(self, scenario):
        if scenario is None:
            return None
        if isinstance(scenario, set):
            match = [scen for scen in filter(None, scenario) if scen in self._anchors.keys()]
            if len(match) == 0:
                return None
            elif len(match) > 1:
                raise ScenarioConflict('fragment: %s\ntermination matches: %s' % (self.uuid, match))
            return match[0]
        if scenario in self._anchors.keys():
            return scenario
        return None

    def anchor(self, scenario):
        s = self._match_scenario_anchor(scenario)
        return self._anchors[s]

    '''
    Balance Flow Handling
    '''
    @property
    def is_balance(self):
        return bool(self._balance)

    @property
    def balance_flow(self):
        return self._balance_child

    def set_balance_flow(self):
        """
        A balance flow balances its own reference quantity.
        :return:
        """
        if self.reference_entity is None:
            raise InvalidParentChild('Reference flow cannot be a balance flow')
        if self.is_balance is False:
            self.reference_entity.set_conservation_child(self)
            self._balance = True

    def unset_balance_flow(self):
        if self.is_balance:
            self.reference_entity.unset_conservation_child()
            self._balance = False

    def set_conservation_child(self, child):
        if child.reference_entity != self:
            raise InvalidParentChild
        if self.is_conserved_parent and (child is not self._balance_child):
            print('%.5s conserving %s\nversus %s' % (self.uuid, self._balance_child, child))
            raise BalanceAlreadySet
        self._balance_child = child
        # if self.balance_magnitude == 0:
        #     logging.warning('%.5s Notice: zero balance for conserved quantity %s' % (self.uuid,
        #                                                                              self.conserved_quantity))
        logging.info('setting balance from %.5s: %s' % (child.uuid, self._balance_child))

    @property
    def is_conserved_parent(self):
        return self._balance_child is not None

    def unset_conservation_child(self):
        self._balance_child = None

    @property
    def conserved_quantity(self):
        if self._balance_child is None:
            return None
        else:
            return self._balance_child.flow.reference_entity

    def clear_anchor(self, scenario=None):
        self._anchors[scenario] = Anchor.null(self)

    def nodes(self, scenario=None, descend=True):
        """
        Report proximal terminal nodes for the fragment (recurse until a nondescend is reached)
        :param scenario: [None]
        :param descend: [True] if False, yield subfragments as nodes
        :return: generator of terminal nodes
        """
        term = self.anchor(scenario)
        yds = set()
        if term.is_process or term.is_context or term.is_unresolved:
            if term.node not in yds:
                yield term.node
                yds.add(term.node)
        elif term.is_subfrag:
            if term.descend and descend:
                for n in term.node.nodes(scenario, descend=descend):
                    if n not in yds:
                        yield n
                        yds.add(n)
            else:
                yield term.node
            # foreground, null: do nothing
        for c in self.child_flows:
            for n in c.nodes(scenario, descend=descend):
                if n not in yds:
                    yield n
                    yds.add(n)

    def tree(self):
        """
        This is real simple- just a recursive enumeration of child flows, depth first

        :return:
        """
        yield self
        for c in sorted(self.child_flows, key=lambda x: (x['StageName'], not x.anch.is_null, x.is_background)):
            for b in c.tree():
                yield b

    def exchanges(self, scenario=None):
        """
        Generator for query compatibility with processes
        :param scenario:
        :return:
        """
        for x in self.child_flows:
            yield x
        for x in self.anchor(scenario).unobserved_exchanges():
            yield x

    def ev(self, scenarios=None):
        """

        :param scenarios: either None, or an iterable of scenarios
        :return: matching scenario, matching ev
        """
        if scenarios is None:
            return self.cached_ev
        else:
            m = self._match_scenario_ev(scenarios)
            if m is None:
                return self.observed_ev
            return self._evs[m]



from antelope.xdb_tokens import ResourceSpec

from .catalog import StaticCatalog, CatalogError
from ..archives import REF_QTYS, archive_from_json
from ..lc_resource import LcResource
from ..lcia_engine import DEFAULT_CONTEXTS, DEFAULT_FLOWABLES
from ..providers.xdb_client.rest_client import RestClient, OAuthToken

import requests
from requests.exceptions import HTTPError

from shutil import copy2, rmtree
import os
import glob
import logging
import hashlib
import getpass

# TEST_ROOT = os.path.join(os.path.dirname(__file__), 'cat-test')  # volatile, inspectable


def download_file(url, local_file, md5sum=None):
    r = requests.get(url, stream=True)
    md5check = hashlib.md5()
    with open(local_file, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
                md5check.update(chunk)
                # f.flush() commented by recommendation from J.F.Sebastian
    if md5sum is not None:
        assert md5check.hexdigest() == md5sum, 'MD5 checksum does not match'


class LcCatalog(StaticCatalog):
    """
    A catalog that supports adding and manipulating resources during runtime
    """
    def download_file(self, url=None, md5sum=None, force=False, localize=True):
        """
        Download a file from a remote location into the catalog and return its local path.  Optionally validate the
        download with an MD5 digest.
        :param url:
        :param md5sum:
        :param force:
        :param localize: whether to return the filename relative to the catalog root
        :return: the full path to the downloaded file 
        """
        if self._test:
            logging.error('Cannot save files locally during tester operation')
            raise CatalogError
        local_file = os.path.join(self._download_dir, self._source_hash_file(url))
        if os.path.exists(local_file):
            if force:
                print('File exists.. re-downloading.')
            else:
                print('File already downloaded.  Force=True to re-download.')
                if localize:
                    return self._localize_source(local_file)
                return local_file

        download_file(url, local_file, md5sum)

        import magic
        if magic.from_file(local_file).startswith('Microsoft Excel 20'):
            new_file = local_file + '.xlsx'
            os.rename(local_file, new_file)  # openpyxl refuses to open files without an extension
            local_file = new_file

        if localize:
            return self._localize_source(local_file)
        return local_file

    @classmethod
    def make_tester(cls, **kwargs):
        """
        This is no longer necessary; the same thing can be accomplished just by calling the constructor with no root.
        :param kwargs:
        :return:
        """
        # tmp = tempfile.mkdtemp()
        return cls(None, _test=True, **kwargs)

    """
    @classmethod
    def load_tester(cls):
        return cls(TEST_ROOT)
    """

    @property
    def _dirs(self):
        for x in (self._cache_dir, self._index_dir, self.resource_dir, self.archive_dir, self._download_dir):
            yield x

    def check_cache(self, source):
        if self._test:
            return False
        return super(LcCatalog, self).check_cache(source)

    def _make_rootdir(self):
        for x in self._dirs:
            os.makedirs(x, exist_ok=True)
        if not os.path.exists(self._contexts):
            copy2(DEFAULT_CONTEXTS, self._contexts)
        if not os.path.exists(self._flowables):
            copy2(DEFAULT_FLOWABLES, self._flowables)
        if not os.path.exists(self._reference_qtys):
            copy2(REF_QTYS, self._reference_qtys)

    def __init__(self, rootdir=None, _test=False, **kwargs):
        if rootdir is None or _test is True:
            self._test = True
            self._rootdir = None
        else:
            self._test = False
            self._rootdir = os.path.abspath(rootdir)
            self._make_rootdir()  # this will be a git clone / fork;; clones reference quantities
        self._blackbook_client = None
        super(LcCatalog, self).__init__(self._rootdir, **kwargs)

    def save_local_changes(self):
        if self._test:
            logging.warning('Cannot save changes during tester operation')
            return
        self._qdb.write_to_file(self._reference_qtys, characterizations=True, values=True)
        self.lcia_engine.save_flowables(self._flowables)
        self.lcia_engine.save_contexts(self._contexts)

    def restore_contexts(self, really=False):
        if self._test:
            logging.warning('Cannot save changes during tester operation')
            return
        if really:
            print('Overwriting local contexts')
            copy2(DEFAULT_CONTEXTS, self._contexts)
        else:
            print('pass really=True if you really want to overwrite local contexts')

    def restore_qdb(self, really=False):
        # this is all deprecated-- need to rework how qdb is initialized and re-initialized
        if self._test:
            logging.info('Cannot save changes during tester operation')
            return
        if really:
            copy2(REF_QTYS, self._reference_qtys)
            print('Reference quantities restored. Please re-initialize the catalog.')

    '''
    Create + Add data resources
    '''

    def new_resource(self, reference, source, ds_type, store=True, **kwargs):
        """
        Create a new data resource by specifying its properties directly to the constructor
        :param reference:
        :param source:
        :param ds_type:
        :param store: [True] permanently store this resource
        :param kwargs: interfaces=None, priority=0, static=False; **kwargs passed to archive constructor
        :return:
        """
        if self._test:
            store = False
        else:
            source = self._localize_source(source)
        res = self._resolver.new_resource(reference, source, ds_type, store=store, **kwargs)
        if res.origin in self._nicknames:
            self._nicknames.pop(res.origin)
        return res

    def add_resource(self, resource, store=True, replace=False):
        """
        Add an existing LcResource to the catalog.
        :param resource:
        :param store: [True] permanently store this resource
        :param replace: [False] if the resource already exists, remove it and replace it with the new resource
        :return:
        """
        if self._test:
            store = False
        if replace:
            for k in self._resolver.matching_resources(resource):
                self.delete_resource(k)
            assert self._resolver.has_resource(resource) is False
        self._resolver.add_resource(resource, store=store)
        if resource.origin in self._nicknames:
            self._nicknames.pop(resource.origin)

    def purge_resource_archive(self, resource: LcResource):
        """
        - find all cached queries that could return the resource
        - check their cached ifaces to see if they use our archive
        - delete those entries from the cache
        :param resource:
        :return:
        """
        # TODO: though this corrects our catalog queries, the entities are not connected to the catalog queries
        for org, q in self._queries.items():
            if resource.origin.startswith(org):
                print('Purging %s' % q)
                q.purge_cache_with(resource.archive)
        print('Removing archive from %s' % resource)
        resource.remove_archive()

    def delete_resource(self, resource, delete_source=None, delete_cache=True):
        """
        Removes the resource from the resolver and also removes the serialization of the resource. Also deletes the
        resource's source (as well as all files in the same directory that start with the same name) under the following
        circumstances:
         (resource is internal AND resources_with_source(resource.source) is empty AND resource.source is a file)
        This can be overridden using he delete_source param (see below)

        We also need to remove any implementations that use the resource.

        :param resource: an LcResource
        :param delete_source: [None] If None, follow default behavior. If True, delete the source even if it is
         not internal (source will not be deleted if other resources refer to it OR if it is not a file). If False,
         do not delete the source.
        :param delete_cache: [True] whether to delete cache files (you could keep them around if you expect to need
         them again and you don't think the contents will have changed)
        :return:
        """
        self._resolver.delete_resource(resource)

        self.purge_resource_archive(resource)

        abs_src = self.abs_path(resource.source)

        if delete_source is False or resource.source is None or not os.path.isfile(abs_src):
            return
        if len([t for t in self._resolver.resources_with_source(resource.source)]) > 0:
            return
        if resource.internal or delete_source:
            if os.path.isdir(abs_src):
                rmtree(abs_src)
            else:
                for path in glob.glob(abs_src + '*'):
                    print('removing %s' % path)
                    os.remove(path)
        if delete_cache:
            if self.check_cache(resource.source):
                os.remove(self.cache_file(resource.source))
            if self.check_cache(abs_src):
                os.remove(self.cache_file(abs_src))

    def add_existing_archive(self, archive, interfaces=None, store=True, **kwargs):
        """
        Makes a resource record out of an existing archive.  by default, saves it in the catalog's resource dir
        :param archive:
        :param interfaces:
        :param store: [True] if False, don't save the record - use it for this session only
        :param kwargs:
        :return:
        """
        res = LcResource.from_archive(archive, interfaces, source=self._localize_source(archive.source), **kwargs)
        self.add_resource(res, store=store)

    def blackbook_authenticate(self, blackbook_url=None, username=None, password=None, token=None, **kwargs):
        """
        Opens an authenticated session with the designated blackbook server.  Credentials can either be provided to the
        method as arguments, or if omitted, they can be obtained through a form.  If a token is provided, it is
        used in lieu of a password workflow
        :param blackbook_url:
        :param username:
        :param password:
        :param token:
        :param kwargs: passed to RestClient. save_credentials=True.  verify: provide path to self-signed certificate
        :return:
        """
        if self._blackbook_client:
            if blackbook_url is None:
                self._blackbook_client.reauthenticate()  # or raise NoCredentials
                return
            self._blackbook_client.close()
        elif blackbook_url is None:
            raise ValueError('Must provide a URL')
        if token is None:
            client = RestClient(blackbook_url, auth_route='auth/token', **kwargs)
            if username is None:
                username = input('Enter username to access blackbook server at %s: ' % blackbook_url)
            if password is None:
                password = getpass.getpass('Enter password to access blackbook server at %s: ' % blackbook_url)
            try:
                client.authenticate(username, password)
            except HTTPError:
                client.close()
                raise
        else:
            client = RestClient(blackbook_url, token=token, auth_route='auth/token', **kwargs)
        self._blackbook_client = client

    def blackbook_guest(self, blackbook_url=None, token=None, **kwargs):
        if self._blackbook_client:
            self._blackbook_client.close()
        client = RestClient(blackbook_url, auth_route='auth/guest', **kwargs)
        if token is None:
            token = client.get_one(OAuthToken, 'auth', 'guest')
        client.set_token(token)
        self._blackbook_client = client

    @property
    def blackbook_origins(self):
        if self._blackbook_client is not None:
            for org in sorted(self._blackbook_client.get_raw('origins')):
                yield org

    def get_blackbook_resources(self, origin, store=False, assign_ref=None, **kwargs):
        """
        Use a blackbook server to obtain resources for a given origin.
        :param origin:
        :param store: whether to save resources. by default we don't, assuming the tokens are short-lived.
        :param assign_ref: give the resource a named ref locally
        :param kwargs: init args to add to returned resources, such as 'verify' certificate paths
        :return:
        """
        if assign_ref is None:
            assign_ref = origin
        res = list(self.resources(assign_ref))
        if len(res) > 0:
            return self.refresh_xdb_tokens(assign_ref)
        else:
            resource_list = self._blackbook_client.get_many(ResourceSpec, 'origins', origin, 'resource')
            for r in resource_list:
                r.options['blackbook_origin'] = r.origin
                r.origin = assign_ref
            return self._configure_blackbook_resources(resource_list, store=store, **kwargs)

    def blackbook_request_third_party_resource(self, origin, resource_for):
        """

        :param origin: the data resource
        :param resource_for: the foreground that needs access
        :return:
        """
        return self._blackbook_client.get_many(ResourceSpec, 'origins', origin, 'resource_for', resource_for)

    def blackbook_reset_tokens(self, foreground):
        return self._blackbook_client.get_one(int, 'origins', foreground, 'refresh_tokens')

    '''
    def get_blackbook_resources_by_client(self, bb_client, username, origin, store=False):
        """
        this uses the local maintenance client rather than the REST client
        :param bb_client:
        :param username:
        :param origin:
        :param store:
        :return:
        """
        resource_dict = bb_client.retrieve_resource(username, origin)
        return self._finish_get_blackbook_resources(resource_dict, store=store)
    '''

    def _configure_blackbook_resources(self, resource_list, store=False, **kwargs):
        """
        Emerging issue here in the xdb/oryx context-- we need to be able to replace resources even if they are
        serialized and already initialized.

        response: this is easy- the XdbClient provider (and subclasses) has refresh_token and refresh_auth methods
        already.

        What this function does: for each entry in resource dict:
         - find the first resource that matches origin + ds_type
         - if one exists, update it:
           = if source matches, update token
           = else, update source and token
         - else: create it

        :param resource_list: a list of [resource specs]
        :return:
        """

        rtn = []

        for res in resource_list:
            if not isinstance(res, ResourceSpec):
                res = ResourceSpec(**res)
            try:
                exis = next(x for x in self.resources(res.origin) if x.ds_type == res.ds_type)
                exis.init_args.update(kwargs)
                exis.check(self)
                # one exists-- update it
                exis.init_args.update(res.options)
                if exis.source == res.source:
                    exis.archive.refresh_token(res.options['token'])
                else:
                    exis._source = res.source
                    exis.archive.refresh_auth(res.source, res.options['token'])
                for i in res.interfaces:
                    if i not in exis.interfaces:
                        exis.add_interface(i)
                for i in exis.interfaces:
                    if i not in res.interfaces:
                        exis.remove_interface(i)
                if store:
                    exis.write_to_file(self.resource_dir)
                rtn.append(exis)
            except StopIteration:
                r = LcResource(**res.model_dump(), **kwargs)
                self.add_resource(r, store=store)
                rtn.append(r)
        return rtn

    def refresh_xdb_tokens(self, remote_origin=None):
        """
        requires an active blackbook client (try blackbook_authenticate() if it has expired)
        :param remote_origin:
        :return:
        """
        rtn = []
        if remote_origin is None:
            for r in self.resources(loaded=True):
                if r.ds_type == 'XdbClient' or r.ds_type == 'OryxClient':
                    rtn.extend(self.refresh_xdb_tokens(r.init_args['blackbook_origin']))
            return rtn

        tok = self._blackbook_client.get_one(OAuthToken, 'origins', remote_origin, 'token')
        for res in self._resolver.resources:
            if res.init_args.get('blackbook_origin') == remote_origin:
                res.init_args['token'] = tok.access_token
                if res.archive is None:
                    res.check(self)
                elif hasattr(res.archive, 'r'):
                    res.archive.r.set_token(tok)
                rtn.append(res)
        return rtn

    '''
    Manage resources locally
     - index
     - cache
     - static archive (performs load_all())
    '''

    def _index_source(self, source, priority, force=False, save=True):
        """
        Instructs the resource to create an index of itself in the specified file; creates a new resource for the
        index
        :param source:
        :param priority:
        :param force:
        :return:
        """
        res = next(r for r in self._resolver.resources_with_source(source))
        res.check(self)
        # priority = min([priority, res.priority])  # we want index to have higher priority i.e. get loaded second
        stored = self._resolver.is_permanent(res) and save

        # save configuration hints in derived index
        cfg = None
        if len(res.config['hints']) > 0:
            cfg = {'hints': res.config['hints']}

        if stored:
            inx_file = self._index_file(source)
            inx_local = self._localize_source(inx_file)

            if os.path.exists(inx_file):
                if not force:
                    print('Not overwriting existing index. force=True to override.')
                    try:
                        ex_res = next(r for r in self._resolver.resources_with_source(inx_local))
                        return ex_res.origin
                    except StopIteration:
                        # index file exists, but no matching resource
                        inx = archive_from_json(inx_file)
                        self.new_resource(inx.ref, inx_local, 'json', priority=priority, store=stored,
                                          interfaces='index', _internal=True, static=True, preload_archive=inx,
                                          config=cfg)

                        return inx.ref

                print('Re-indexing %s' % source)
                stale_res = list(self._resolver.resources_with_source(inx_local))
                stale_refs = list(set(res.origin for res in stale_res))
                for stale in stale_res:
                    # this should be postponed to after creation of new, but that fails in case of naming collision
                    # (bc YYYYMMDD)
                    # so golly gee we just delete-first.
                    print('deleting %s' % stale.origin)
                    self.delete_resource(stale)
                # we also need to delete derived internal resources
                for stale_ref in stale_refs:
                    for stale in list(self.resources(stale_ref)):
                        if stale.internal:
                            self.delete_resource(stale)
        else:
            inx_file = inx_local = None

        the_index = res.make_index(inx_file, force=force, save=stored)
        if inx_local is None:
            inx_local = the_index.ref
        nr = self.new_resource(the_index.ref, inx_local, 'json', priority=priority, store=stored,
                               interfaces=('basic', 'index'),
                               _internal=True, static=True, preload_archive=the_index, config=cfg)
        if nr.priority > res.priority:
            # this allows the index to act to retrieve entities if the primary resource fails
            nr.add_interface('basic')

        return the_index.ref

    def index_ref(self, origin, interface=None, source=None, priority=60, save=True, force=False, strict=True):
        """
        Creates an index for the specified resource.  'origin' and 'interface' must resolve to one or more LcResources
        that all have the same source specification.  That source archive gets indexed, and index resources are created
        for all the LcResources that were returned.

        Performs load_all() on the source archive, writes the archive to a compressed json file in the local index
        directory, and creates a new LcResource pointing to the JSON file.   Aborts if the index file already exists
        (override with force=True).
        :param origin:
        :param interface: [None]
        :param source: find_single_source input
        :param priority: [60] priority setting for the new index -- authentic source is highest
        :param save: [True] whether to save the index
        :param force: [False] if True, overwrite existing index
        :param strict: [True] whether to be strict
        :return:
        """
        if not force:
            try:
                ix = next(self.gen_interfaces(origin, itype='index', strict=False))
                return ix.origin
            except StopIteration:
                pass
        source = self._find_single_source(origin, interface, source=source, strict=strict)
        return self._index_source(source, priority, force=force, save=save)

    def cache_ref(self, origin, interface=None, source=None, static=False):
        source = self._find_single_source(origin, interface, source=source)
        self.create_source_cache(source, static=static)

    def create_source_cache(self, source, static=False):
        """
        Creates a cache of the named source's current contents, to speed up access to commonly used entities.
        source must be either a key present in self.sources, or a name or nickname found in self.names
        :param source:
        :param static: [False] create archives of a static archive (use to force archival of a complete database)
        :return:
        """
        res = next(r for r in self._resolver.resources_with_source(source))
        if res.static:
            if not static:
                print('Not archiving static resource %s' % res)
                return
            print('Archiving static resource %s' % res)
        res.check(self)
        res.make_cache(self.cache_file(self._localize_source(source)))

    def _background_for_origin(self, ref, strict=False):
        res = self.get_resource(ref, iface='exchange')
        store = self._resolver.is_permanent(res) and not self._test
        inx_ref = self.index_ref(ref, interface='exchange', strict=strict, save=store)
        if store:
            bk_file = self._localize_source(os.path.join(self.archive_dir, '%s_background.mat' % inx_ref))
        else:
            bk_file = '%s_background.mat' % inx_ref
        bk = LcResource(inx_ref, bk_file, 'Background', interfaces='background', priority=99,
                        save_after=store, _internal=True)
        bk.config = res.config
        bk.check(self)  # ImportError if antelope_background pkg not found;; also applies configs
        self.add_resource(bk, store=store)
        return bk.make_interface('background')  # when the interface is returned, it will trigger setup_bm

    def gen_interfaces(self, origin, itype=None, strict=False, ):
        """
        Override parent method to also create local backgrounds
        :param origin:
        :param itype:
        :param strict:
        :return:
        """
        for k in super(LcCatalog, self).gen_interfaces(origin, itype=itype, strict=strict):
            yield k

        if itype == 'background':
            if origin.startswith('local') or origin.startswith('test'):
                yield self._background_for_origin(origin, strict=strict)

    def create_descendant(self, origin, interface=None, source=None, force=False, signifier=None, strict=True,
                          priority=None, **kwargs):
        """

        :param origin:
        :param interface:
        :param source:
        :param force: overwrite if exists
        :param signifier: semantic descriptor for the new descendant (optional)
        :param strict:
        :param priority:
        :param kwargs:
        :return:
        """
        res = self.get_resource(origin, iface=interface, source=source, strict=strict)
        new_ref = res.archive.create_descendant(self.archive_dir, signifier=signifier, force=force)
        print('Created archive with reference %s' % new_ref)
        ar = res.archive
        prio = priority or res.priority
        self.add_existing_archive(ar, interfaces=res.interfaces, priority=prio, **kwargs)
        res.remove_archive()



"""
Client for the xdb Antelope server

The concept here is to pass received requests straight into the Pydantic models, and then use those for easy
(though manual) deserialization into EntityRefs.
"""
from antelope import EntityNotFound
from antelope.models import Context as ContextModel, Entity
from antelope_core.archives import LcArchive, InterfaceError
from antelope_core.catalog_query import READONLY_INTERFACE_TYPES
from antelope_core.contexts import ContextManager, NullContext

from .requester import XdbRequester
from .implementation import XdbImplementation, XdbConfigureImplementation, _ref
from .xdb_entities import XdbEntity

from requests.exceptions import HTTPError


class XdbTermManager(object):
    def __init__(self, requester: XdbRequester):
        """
        The key question here: should I at least cache remote contexts? for now let us cache nothing

        Except we DO need to cache contexts, or at least create them on the fly (and if we are going to create
        them, we should cache them) because exterior exchanges need to terminate properly in contexts.

        SO: we will establish the condition that Xdb MUST only use contexts' canonical names, thereby escaping the
        need for synonym disambiguation in the client.
        :param requester:
        """
        self._requester = requester
        self._cm = ContextManager()
        self._bad_contexts = set()
        self._flows = set()
        self._quantities = set()

    def update_requester(self, requester):
        self._requester = requester

    @property
    def is_lcia_engine(self):
        return self._requester.is_lcia_engine

    def _fetch_context_model(self, item):
        return self._requester.get_one(ContextModel, 'contexts', item)

    def _build_context(self, context_model, *args):
        if context_model.name == 'None':
            c_actual = NullContext  # maintain singleton status of NullContext
        else:
            if context_model.parent is None or context_model.parent == '':
                parent = None
            else:
                parent = self.get_context(context_model.parent)
            c_actual = self._cm.new_entry(context_model.name, *args, parent=parent)
            if parent is not None and parent.sense is None and context_model.sense is not None:
                c_actual.sense = context_model.sense
            c_actual.add_origin(self._requester.origin)  # here we are masquerading all origins back to the requester origin
        return c_actual

    def add_flow(self, flow, **kwargs):
        self._flows.add(flow)

    def add_quantity(self, quantity):
        self._quantities.add(quantity)

    def __getitem__(self, item):
        if isinstance(item, list):
            item = tuple(item)
        if item in self._bad_contexts:
            return None
        try:
            return self._cm[item]
        except KeyError:
            try:
                c_model = self._fetch_context_model(item)
            except HTTPError as e:
                if e.args[0] == 404:
                    self._bad_contexts.add(item)
                    return None
                else:
                    raise
            c_actual = self._build_context(c_model, item)
            # this escapes the PROTECTED_TERM restrictions- which is OK because our CM is captive
            self._cm.add_synonym(c_actual.name, item)
            return c_actual

    def is_context(self, item):
        """
        The only place this is used is in collision checking- therefore this only needs to check if the name is
        known *locally* as a context (why do an http query for literally every new entity?)
        :param item:
        :return:
        """
        try:
            cx = self._cm[item]
        except KeyError:
            return False
        return cx is not None

    def get_context(self, item):
        return self.__getitem__(item) or NullContext

    '''
    def get_canonical(self, item):
        """
        again, to avoid premature optimization, the initial policy is no caching anything
        :param item:
        :return:
        """
        try:
            return self._requester.get_one(Entity, 'quantities', item)
        except HTTPError as e:
            if e.args[0] == 404:
                raise EntityNotFound(item)
            else:
                raise e
    '''

    def synonyms(self, term):
        return self._requester.get_many(str, 'synonyms', term=term)

    def contexts(self, **kwargs):
        c_models = self._requester.get_many(ContextModel, 'contexts', **kwargs)
        for c in c_models:
            if c.name in self._cm:
                yield self._cm[c.name]
            else:
                yield self._build_context(c)


class XdbClient(LcArchive):
    """
    An XdbClient accesses xdb at a named URL using an access token.
    """

    _base_type = XdbEntity

    def __init__(self, source, ref=None, token=None, blackbook_origin=None, **requester_args):
        self._requester_args = requester_args
        if blackbook_origin is None:
            blackbook_origin = ref
        self._blackbook_origin = blackbook_origin
        try:
            self._requester = XdbRequester(source, self._blackbook_origin, token=token, **self._requester_args)
        except HTTPError as e:
            raise InterfaceError('HTTP Request failed %s, %s' % (e.args[0], e.args[1]))
        if ref is None:
            ref = 'qdb'
        super(XdbClient, self).__init__(source, ref=ref, term_manager=XdbTermManager(self._requester))

    def refresh_token(self, new_token):
        self._requester.set_token(new_token)

    def refresh_auth(self, new_source, new_token):
        self._requester = XdbRequester(new_source, self._blackbook_origin, token=new_token, **self._requester_args)
        self.tm.update_requester(self._requester)

    @property
    def r(self):
        return self._requester

    def make_interface(self, iface):
        if iface in READONLY_INTERFACE_TYPES:
            return XdbImplementation(self)
        elif iface == 'configure':
            return XdbConfigureImplementation(self)
        raise InterfaceError(iface)

    def _model_to_entity(self, model):
        model.properties['blackbook_origin'] = model.origin
        if model.origin == self._blackbook_origin:
            model.origin = self.ref  # yet another masquerade
        entity = self._base_type(model)
        self._entities[model.link] = entity
        return entity

    def get_or_make(self, model):
        """
        Retrieve or create an entity, when we have already received its model data from the server
        :param model:
        :return:
        """
        key = model.link
        if key in self._entities:
            return self._entities[key]
        return self._model_to_entity(model)

    def __getitem__(self, item):
        """
        For cursed reasons, our EntityStore __getitem__ method must return None instead of raising a KeyError
        :param item:
        :return:
        """
        try:
            if hasattr(item, 'link'):
                return self._entities[item.link]
            return self._entities[item]
        except KeyError:
            return None

    def _fetch(self, key, origin=None, **kwargs):
        # I'm just reimplementing _ref_to_key except in situ  ## EntityStore is whack legacy
        if hasattr(key, 'link'):
            link = key.link
            origin = origin or key.origin
            key = key.external_ref
        else:  # anything that has "external_ref" would have "link", right?
            if origin:
                link = '/'.join([origin, str(key)])
            else:
                link = '/'.join([self.ref, str(key)])
        if link in self._entities:
            return self._entities[link]
        try:
            if origin:
                model = self._requester.origin_get_one(Entity, origin, _ref(key))
            else:
                model = self._requester.get_one(Entity, _ref(key))
        except HTTPError as e:
            if e.args[0] == 404:
                raise EntityNotFound(_ref(key))
            else:
                raise
        return self._model_to_entity(model)



"""
autorange.py

Scales numerical values by powers of 1000 to present engineering notation. Also scales metric prefixes to match.
"""
from math import log10


metric_offsets = {
    'a': -6,
    'f': -5,
    'p': -4,
    'n': -3,
    'u': -2,
    'm': -1,
    'k': 1,
    'M': 2,
    'G': 3,
    'T': 4,
    'P': 5,
    'E': 6,
    'Z': 7
}


numerals = ['', 'trillion', 'billion', 'million', 'thousand']  # 0 = nothing -1 = thousand


metric_prefixes = dict([(v, k) for k, v in metric_offsets.items()])
metric_prefixes[0] = ''


DISALLOW = {'mol', 'MT', 'PM'}  # units that start with these phrases should NOT use inferred prefix


class AutoRange(object):
    def _set_shift(self):
        self._shift = 0
        r = log10(self._range)
        while True:
            if r < self.bias:
                r += 3
                self._shift += 1
            elif r > 3 + self.bias:
                r -= 3
                self._shift -= 1
            else:
                return

    def __init__(self, rng, bias=0, kg_to_t=True, disallow_prefix=None):
        """
        creates an object for computing auto-ranged values.  The input argument should be the largest value that
        is expected to appear in the context.  It will be ranged to fall between 0-1000 (absolute value); this
        can be adjusted with the 'bias' param: one order per integer value up (pos) or down (neg)
        :param rng: either a scalar (abs) or an iterable (max(abs))
        :param bias: [0] default range is 0-1000.  bias=1: 10-10,000; bias=-1: 0.1-100; etc; floats are operable
        :param kg_to_t: [True] correct kg to t in cases where the results are scaled up
        example: with kg_to_t = True, a unit of 'kg' autoranged to 'Mg' or larger will be converted to 't',
        'Gg' to 'kt', etc, but a unit of 'kg' or smaller will not
        with kg_to_t False, no alteration is performed
        :param disallow_prefix: preifx(es) to add to DISALLOW ('MT', 'mol', 'PM'). can be string or iterable.
        """
        self.bias = bias
        self.disallow_prefix = set(DISALLOW)
        if disallow_prefix:
            if isinstance(disallow_prefix, str):
                self.disallow_prefix.add(disallow_prefix)
            else:
                for d in disallow_prefix:
                    self.disallow_prefix.add(d)

        try:
            self._range = max(abs(k) for k in rng)
        except TypeError:
            self._range = abs(rng)
        if self._range == 0:
            print('Warning: autorange initialized with 0 range; using 1')
            self._range = 1
        self._shift = 0
        self._set_shift()
        self.kg_to_t = kg_to_t

    @property
    def kg_to_t(self):
        return self._kg_to_t

    @kg_to_t.setter
    def kg_to_t(self, value):
        self._kg_to_t = bool(value)

    @property
    def scale(self):
        return 10 ** (3 * self._shift)

    def adjust(self, value):
        return value * self.scale

    def disallow(self, prefix):
        self.disallow_prefix.add(prefix)
        return self.disallow_prefix

    @property
    def numeral(self):
        """
        report scaling factor as multiplicative numeral
        :return:
        """
        if self._shift < 0:
            return numerals[self._shift]  # counting from end
        return ''

    def adj_unit(self, unit):
        """
        This is the tricky, ad-hoc part.  Assume the first character in the unit string is a prefix IF it is found
        in the metric prefixes dictionary.  If it is not, assume it is unprefixed.
        :param unit: a unit string
        :return:
        """
        prefix = unit[0]

        if any([unit.startswith(x) for x in self.disallow_prefix]) or prefix not in metric_offsets:
            pre_shift = 0
            adj = unit
        else:
            pre_shift = metric_offsets[prefix]
            adj = unit[1:]

        post_shift = pre_shift - self._shift

        if unit.startswith('kg') and post_shift >= 2 and self.kg_to_t:
            post_shift -= 2
            adj = 't' + unit[2:]

        return metric_prefixes[post_shift] + adj



import ast
from collections import deque
from synonym_dict import LowerDict  # apparently Andreas does not check case

import logging

from .schema_mapping import OLCA_MAPPING
import re


class NodeIdentifier(ast.NodeVisitor):
    """
    identifies all the parameters named in an expression
    """
    def __init__(self):
        self.names = list()

    def visit_Name(self, node: ast.Name):
        if node.id not in self.names:
            self.names.append(node.id)
        self.generic_visit(node)


class NodeCalculator(ast.NodeTransformer):
    """
    substitutes constant values in for parameter names, and evaluates the expression
    """
    def __init__(self, values):
        self.values = values

    def visit_Name(self, node: ast.Name):
        if node.id in self.values:
            new_node = ast.Constant(self.values[node.id])
            return ast.copy_location(new_node, node)
        return node


class _Param(object):
    """
    Creates a param from a dictionary spec.
    At this level, the only required dictionary key is 'name'. optional 'unit' can also be overwritten.
    """
    _mkr = '_#'
    _unit = None

    def __init__(self, param):
        self._param = param

    @property
    def name(self):
        return self._param['name'].strip()

    @property
    def unit(self):
        if self._unit:
            return self._unit
        return self._param.get('unit')

    @unit.setter
    def unit(self, value):
        if self._unit:
            if value is None:
                self._unit = None
            elif str(value) != self._unit:
                raise ValueError('Unit %s conflicts with existing value %s' % (value, self._unit))
        else:
            self._unit = str(value)

    @property
    def level(self):
        return 0

    @property
    def dependencies(self):
        yield self.name

    def __str__(self):
        return '%s[%02d] %s' % (self._mkr, self.level, self.name)


class ParamConstant(_Param):
    """
    An input.  'param' API:
    dict. keys: 'name', 'value' (a float), 'unit' (an optional string)
    """
    @property
    def _mkr(self):
        if self._set is None:
            return ' *'
        return ' $'

    _set = None

    def set(self, value):
        self._set = float(value)

    def unset(self):
        self._set = None

    @property
    def value(self):
        if self._set is not None:
            return self._set
        return self._param['value']


class ConditionalParam(object):
    """
    Evaluate an expression, and return a if true, b if false
    this is nuts--- they're even nested.
    ChatGPT tells me I could use simple regexes to change e.g. IF(expr1; expr2; expr3) to (expr2 if expr1 else expr3)
    but trying to handle nested expressions with regexp seems foolhardy.

    I mean, look at this:
    IF(auto_calc_wtr=1;IF(0.474454781149916-(pct_fresh_srf-0.167302119871893)*(0.474454781149916*1/(0.358243098978191+0.474454781149916))>0.0424528301886793;0.474454781149916-(pct_fresh_srf-0.167302119871893)*(0.474454781149916*1/(0.358243098978191+0.474454781149916));0.0424528301886793);usr_fresh_gnd)

    """
    def __init__(self, name, conditional, if_true, if_false):
        self._name = name
        self._cond = conditional
        self._if_true = if_true
        self._if_false = if_false


class FormulaParser(_Param):
    """
    give it a parameter spec; it will act as a dynamic calculator.
    'param' api: dict with keys 'name' (str), 'value' (float), 'formula' (str with elementary operations + groups)
    """
    _mkr = '  '
    _faulty = False

    def __init__(self, engine, param):
        super(FormulaParser, self).__init__(param)
        self._engine = engine
        self._ni = NodeIdentifier()
        try:
            self._ni.visit(ast.parse(self.formula, mode='eval'))
        except SyntaxError:
            logging.error('(%s) %s - Unsupported formula\n%s' % (self._engine.process_ref,
                                                                 self.name,
                                                                 self.formula))
            self._faulty = True
            self._mkr = 'XX'

    @property
    def formula(self):
        if self._faulty:
            return str(self._param['value'])
        form = self._param['formula']
        if form.find('^') > 0:
            pattern = r"(\s|\b)(\^)(\s|\b)"
            form = re.sub(pattern, r'\1**\3', form)
        return form

    @property
    def variables(self):
        for k in self._ni.names:
            yield k

    @property
    def dependencies(self):
        """
        Generates input variables that are used to calculate the parameter value
        :return:
        """
        ys = set()
        for k in self.variables:
            for dk in self._engine[k].dependencies:
                if dk not in ys:
                    yield dk
                    ys.add(dk)

    @property
    def value(self):
        d = {k: self._engine.value(k) for k in self.variables}
        nc = NodeCalculator(d)
        transformed = nc.visit(ast.parse(self.formula, mode='eval'))
        c = compile(transformed, filename="foo", mode='eval')
        result = eval(c)
        if self._engine.noisy:
            print('%.5g %s' % (result, self))
        return result

    @property
    def level(self):
        return max(self._engine[k].level for k in self.variables) + 1


class OlcaParameterResolver(object):
    """
    A class to handle and compute parameters in the OLCA framework.
    Ultimately we want to be able to represent computed parameters involving sums as balance flows
    (and products as child flows)
    """
    def __init__(self, process_json, noisy=False, v2=True, process_ref=None):
        """
        Supply the json object.  will:
         - extract input parameters
         - extract and order derived parameters
         - confirm parameter values

        :param process_json:
        :param noisy: [False] print extensive debugging info
        :param v2: [True] whether to use the v2 schema or not
        """
        _input_param = 'inputParameter'

        if v2:
            _input_param = OLCA_MAPPING['Parameter'][_input_param]

        self._noisy = bool(noisy)

        self._pj = process_json
        self._process_ref = process_ref
        self._params = process_json.get('parameters', ())
        self._inputs = [ParamConstant(p) for p in self._params if bool(p.get(_input_param, False))]
        self._formulas = [FormulaParser(self, p) for p in self._params if not bool(p.get(_input_param, False))]

        self._map = LowerDict((p.name, p) for p in self._inputs)
        for f in self._formulas:
            self._map[f.name] = f
        self._xp = {x['internalId']: x['amountFormula'] for x in process_json['exchanges'] if x.get('amountFormula')}
        self._xs = {x['internalId']: x['amount'] for x in process_json['exchanges']}

        # try and set units, noting that OLCA schema does not enforce consistency
        for x in process_json['exchanges']:
            formula = x.get('amountFormula')
            if formula:
                try:
                    p = self._map[formula]
                except KeyError:
                    adhoc = {'name': formula, 'value': x['amount'], 'formula': formula}
                    p = FormulaParser(self, adhoc)
                    self._formulas.append(p)
                    self._map[p.name] = p
                try:
                    p.unit = x['unit']['name']
                except ValueError:
                    logging.warning('Conflicting units for %s (set: %s, attempted: %s)' % (x['amountFormula'],
                                                                                           p.unit,
                                                                                           x['unit']['name']))

        self._ordered = []
        self._order_params()

        self._stack = []

    @property
    def process_ref(self):
        return str(self._process_ref)

    @property
    def active(self):
        return len(self._params) > 0

    @property
    def noisy(self):
        return self._noisy

    @noisy.setter
    def noisy(self, value):
        self._noisy = bool(value)

    def _order_params(self):
        # put the inputs at the front
        self._ordered = list(self.inputs)
        dq = deque(k for k in self.formulas)
        count = 0
        while dq:
            g = dq.popleft()
            fg = self._map[g]
            if all(k in self._ordered for k in fg.variables):
                self._ordered.append(g)
                count = 0
            else:
                dq.append(g)
                count += 1
                if count > len(dq):
                    raise RecursionError('too many tries %d')

    @property
    def variables(self):
        for k in self._ordered:
            yield k

    @property
    def params(self):
        for k in self._ordered:
            yield self[k]

    @property
    def inputs(self):
        for k in self._inputs:
            yield k.name

    @property
    def formulas(self):
        for k in self._formulas:
            yield k.name

    @property
    def outputs(self):
        """
        parameters that are linked to an exchange
``        :return:
        """
        for v in self._xp.values():
            yield v

    def __getitem__(self, item):
        return self._map[item.strip()]

    def __contains__(self, item):
        return self._map.__contains__(item)

    def set_value(self, key, value):
        param = self._map[key]
        if param in self._inputs:
            param.set(value)
        else:
            raise TypeError('Cannot set derived parameter %s' % key)

    def unset_value(self, key):
        param = self._map[key]
        if param in self._inputs:
            param.unset()
        else:
            raise TypeError('Cannot set derived parameter %s' % key)

    def set_values(self, **kwargs):
        for k, v in kwargs.items():
            self.set_value(k, v)

    def unset_values(self, *args):
        for k in args:
            self.unset_value(k)

    def value(self, key):
        p = self._map[key]
        if self._noisy:
            print(p)
        if key in self.inputs:
            return p.value
        else:
            if key in self._stack:
                raise RecursionError('key %s already visited' % key)
            self._stack.append(key)
            v = p.value  # this recurses
            self._stack.remove(key)
            return v

    def exchange_value(self, internal_id):
        try:
            return self.value(self._xp[internal_id])
        except KeyError:
            return self._xs[internal_id]

    def param_unit(self, param):
        """
        This is inaccurate because the same param can be used for multiple exchanges regardless of their flow
        property. This will only capture the units applied to ONE USE of the parameter name (nondeterministically)
        :param param:
        :return:
        """
        return self._map[param].unit



import csv
import os

from pydantic import BaseModel
from typing import Dict, Optional

from zipfile import ZipFile
from io import TextIOWrapper


class UnitGroup(BaseModel):
    name: str
    ref_unit: str
    conversion: Dict[str, float]  # my funky inverted UnitConversion dictionaries
    # (where every entry corresponds to a unit of the reference quantity)


class Location(BaseModel):
    uuid: str
    name: str
    description: Optional[str]
    code: str
    latitude: float
    longitude: float

    @classmethod
    def from_csv_dict(cls, record):
        return cls(uuid=record['ID'], name=record['Name'], description=record['Description'], code=record['Code'],
                   latitude=float(record['Latitude']), longitude=float(record['Longitude']))


class OlcaAccessorV2(object):
    """
    To access the reformed GreenDelta/data repository
    """
    def __init__(self, olca_root):
        self._root = olca_root
        if os.path.isdir(olca_root):
            self._zip = None
        else:
            self._zip = ZipFile(olca_root)

    @property
    def refdata(self):
        if self._zip:
            return 'refdata'
        return os.path.join(self._root, 'refdata')

    @property
    def lcia_factors(self):
        return os.path.join(self.refdata, 'lcia_factors')

    def get_refdata(self, ref_file):
        f = os.path.join(self.refdata, ref_file)
        if not f.endswith('.csv'):
            f += '.csv'
        if self._zip:
            with self._zip.open(f) as fp:
                wrapper = TextIOWrapper(fp, encoding='utf-8', newline='')
                d_r = csv.DictReader(wrapper, delimiter=",")
                data = list(d_r)
        else:
            with open(f) as fp:
                d_r = csv.DictReader(fp, delimiter=",")
                data = list(d_r)
        return data

    def read_unit_groups(self):
        """
        This returns a DICT whose KEYS are the Unit Group Name, and whose VALUES are UnitGroup objects
        That dict is then accessed later on
        :return:
        """
        ugs = self.get_refdata('unit_groups')
        us = self.get_refdata('units')

        unit_groups = {ug['Name']: UnitGroup(name=ug['Name'], ref_unit=ug['Reference unit'], conversion=dict())
                       for ug in ugs}

        # first we construct the U/C dicts as they are specified
        for u in us:
            ug = unit_groups[u['Unit group']]
            factor = float(u['Conversion factor'])
            ug.conversion[u['Name']] = factor
            for syn in u['Synonyms'].split(';'):
                ug.conversion[syn] = factor

        # then we invert them to be funky like ours
        for ug in unit_groups.values():
            ref = ug.conversion[ug.ref_unit]
            for k in list(ug.conversion.keys()):
                ug.conversion[k] = ref / ug.conversion.pop(k)

        return unit_groups

    def read_locales(self):
        """
        locations.csv has ID,Name,Description,Category,Code,Latitude,Longitude columns, but:
        - "Category" is empty for all records
        - "ID" we're skipping
        - "Latitude" and "Longitude" we have no use for at the moment
        - and we don't need "Description" either. but let's just keep em all.
        :return:
        """
        ls = self.get_refdata('locations')
        return {l['Code']: Location.from_csv_dict(l) for l in ls}

    def collate_lcia_categories(self):
        dd = dict()
        for row in self.get_refdata('lcia_method_categories'):
            if row['LCIA method'] not in dd:
                dd[row['LCIA method']] = []
            dd[row['LCIA method']].append(row['LCIA category'])
        return dd

    def factors_for_method(self, q):
        """
        LCIA category,Flow,Flow property,Flow unit,Location,Factor
        :param q:
        :return:
        """
        f = None
        i = 0
        l_count = 0
        while f is None:
            i += 1
            cand = os.path.join(self.lcia_factors, q.uuid[:i] + '.csv')
            try:
                self._zip.getinfo(cand)
                f = cand
            except KeyError:
                pass
            if i > 10:
                raise FileNotFoundError(cand)

        if self._zip:
            with self._zip.open(f) as fp:
                wrapper = TextIOWrapper(fp, encoding='utf-8', newline='')
                d_r = csv.DictReader(wrapper, delimiter=",")
                for row in d_r:
                    if row['Location']:
                        l_count += 1
                    yield row
        else:
            with open(f) as fp:
                d_r = csv.DictReader(fp, delimiter=",")
                for row in d_r:
                    if row['Location']:
                        l_count += 1
                    yield row

        if l_count:
            print('method %s: %d non-null Locations' % (q.uuid, l_count))



from setuptools import setup, find_packages

VERSION = '0.3.4.1'

requires = [
    "synonym_dict>=0.2.4",
    "antelope_interface>=0.2.7",
    "xlstools>=0.1.3",
    "python-magic>=0.4.18",
    "requests>=2.25",
    "pydantic>=2.5.0"
]

# optional: pylzma
"""
Version History
0.3.4.1 2024-10-09 - fixed a flowables issue: PM2.5 and PM10 had become merged.

0.3.4  2024-09-22  - workshop release.  Ditch make_tester()

0.3.4-pre 2024-09-22 - workshop pre-release. fix lots of wee bugs; add docs

0.3.3   2024-08-10 - consumers implementation
                     sorting direction on LciaResults
                     clean up unobserved_lci in preparation for folding it into lci()
                     catch some errors
                     some new unit conversions
                     
0.3.2.3 2024-08-15 - lots of remote traversal compatibility changes
                     StringEntities in LciaResults
                     repair biogenic CO2 switching
                     ecoinvent 3.10 flowables
                     assign_ref for blackbook resources
                     zillions of tiny, costly bugs

0.3.2.2 2024-06-11 - get_context into basic interface
0.3.2   2024-06-06 - create ItemNotFound - bump interface version

0.3.1   2024-04-18 - update antelope-interface to 0.2.4

0.3.0   2024-01-05 - 0.3-branch development version, supporting end-user access to vault.lc resources 

# ^ 0.3.* 0.3-branch fork   
# v 0.2.* main / master for legacy projects

0.2.5   2024-05-06 - move bg_lcia() into basic interface; sys_lcia() as fully-featured background analogue

0.2.4.1 2024-04-24 - implement pagination in BasicArchive search
                     fix ecospold v1 to handle process-flow collisions
                     add_interfaces in LcResource
                     LcCatalog.add_resource(replace=True)

0.2.4   2024-04-17 - Remove antelope.ExteriorFlow in favor of antelope.models.ExteriorFlow

0.2.3   2024-03-21 - compatibility release for antelope_interface upgrades and terminology changes
                     OpenLCA v2 schema handled
                     lots of cloud work 

0.2.1   2023-04-10 - xdb passes benchmarks
                     pydantic models moved into interface
                     sys_lci and bg_lcia operational, both locally and remotely

0.2.0   2023-04-07 - "release" virtualize branch. 
                   - Add pydantic models for everything
                   - add an XDB client implementation for remote operation
                   - job-related changes to entity handling throughout

0.1.8   2022-04-08 - PyPI release to roll up a few small changes:
 * upgrade ecoinvent_lcia to use xlstools; port 3.8
 - pull out parse_math
 - windows path mgmt
 - background trivial impl needs to access entities and not refs
 - LciaResults can sum together if at least one has a unit node weight
 - cleanup reference quantities; finally add node activity as canonical quantity
 - quantity unit mismatch refactor


0.1.7.3 2021-08-11 - fix last edit
0.1.7.2 2021-08-11 - make lxml optional when running _find_providers
0.1.7.1 2021-08-11 - eliminate recursion error in basic get_reference()
0.1.7 - 2021-08-05 - merge configuration changes from [virtualize] in-progress
0.1.6 - 2021-03-10 - update to handle new synonym_dict 0.2.0, along with OLCA reference flow matching, ecoinvent 2.2,
                     a range of other improvements in performance and context handling
                     2021-03-10 post1
                     2021-03-17 post2 - get_context(); bump antelope version
                     
0.1.5 - 2021-02-05 - Updates to NullContext handling, flow term matching, fixed faulty requirements (add requests) 
0.1.4 - 2021/01/29 - background passing
0.1.3 - 2021/01/29 - Build passing (without bg tested)
0.1.3rc4           - fix last edit
0.1.3rc3           - xlrd removes support for xlsx in 2.0; _localize_file(None) return None
0.1.3rc2           - include data files
0.1.3rc1           - update requirements

0.1.2b - 2020/12/29 - fix some minor items
0.1.2 - 2020/12/28 - PyPI installation; includes significant performance enhancements for LCIA 

0.1.1 - 2020/11/12 - Bug fixes all over the place.  
                     Catalogs implemented
                     LCIA computation + flat LCIA computation reworked
                     Improvements for OpenLCA LCIA methods

0.1.0 - 2020/07/31 - Initial release - JIE paper
"""

setup(
    name="antelope_core",
    version=VERSION,
    author="Brandon Kuczenski",
    author_email="bkuczenski@ucsb.edu",
    license="BSD 3-Clause",
    install_requires=requires,
    extras_require={
        'XML': ['lxml>=1.2.0'],
        'write_to_excel': ['xlsxwriter>=1.3.7']
    },
    include_package_data=True,
    url="https://github.com/AntelopeLCA/core",
    summary="A reference implementation of the Antelope interface for accessing a variety of LCA data sources",
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
    packages=find_packages()
)



"""
Query Interface -- used to operate catalog refs
"""
import logging

from antelope import (BasicInterface, IndexInterface, BackgroundInterface, ExchangeInterface, QuantityInterface,
                      EntityNotFound, UnknownOrigin,
                      ExchangeRef, comp_dir, CatalogRef)

from antelope.interfaces.ibasic import BasicRequired

from antelope.refs import FlowRef, RxRef

from antelope.models import LciaResult as LciaResultModel, Characterization as CharacterizationModel

from .lcia_results import LciaResult, MixedComponents
from .characterizations import QRResult, Characterization
from .contexts import NullContext
from .implementations.quantity import CO2QuantityConversion

from .fragment_flows import FragmentFlow, frag_flow_lcia

from synonym_dict import InconsistentLineage

INTERFACE_TYPES = ('basic', 'index', 'exchange', 'background', 'quantity', 'foreground')
READONLY_INTERFACE_TYPES = {'basic', 'index', 'exchange', 'background', 'quantity'}


def zap_inventory(interface, warn=False):
    if interface == 'inventory':
        if warn:
            print('# # # # # # # # # **** Warning: use exchange over inventory ***** # # # # # # # # #')
            raise AttributeError
        return 'exchange'
    return interface


class NoCatalog(Exception):
    pass


class BackgroundSetup(Exception):
    pass


class BadInterfaceSpec(Exception):
    pass


class CatalogQuery(BasicInterface, IndexInterface, BackgroundInterface, ExchangeInterface, QuantityInterface):
    """
    A CatalogQuery is a class that performs any supported query against a supplied catalog.
    Supported queries are defined in the lcatools.interfaces, which are all abstract.
    Implementations also subclass the abstract classes.

    This reduces code duplication (all the catalog needs to do is provide interfaces) and ensures consistent signatures.

    The arguments to a query should always be text strings, not entities.  When in doubt, use the external_ref.

    The EXCEPTION is the bg_lcia routine, which works best (auto-loads characterization factors) if the query quantity
    is a catalog ref.

    The catalog's resolver performs fuzzy matching, meaning that a generic query (such as 'local.ecoinvent') will return
    both exact resources and resources with greater semantic specificity (such as 'local.ecoinvent.3.2.apos').
    All queries accept the "strict=" keyword: set to True to only accept exact matches.
    """
    _recursing = False
    _dbg = False

    def on_debug(self):
        self._dbg = True

    def off_debug(self):
        self._dbg = False

    def _debug(self, *args):
        if self._dbg:
            print(self.__class__.__name__, *args)

    def __init__(self, origin, catalog=None, debug=False):
        self._origin = origin
        self._catalog = catalog
        self._dbg = debug

        self._iface_cache = dict()

    def __str__(self):
        if self._catalog:
            root = 'catalog_root=%s' % self._catalog.root
        else:
            root = 'no catalog'
        if self._dbg:
            root += ', DEBUG ON'
        return '%s(%s, %s)' % (self.__class__.__name__, self._origin, root)

    def __repr__(self):
        return self.__str__()

    def purge_cache_with(self, archive):
        for i, v in list(self._iface_cache.items()):
            if v.is_me(archive):
                self._iface_cache.pop(i)

    @property
    def origin(self):
        return self._origin

    @property
    def _tm(self):
        return self._catalog.lcia_engine

    '''
    def is_elementary(self, context):
        """
        Stopgap used to expose access to a catalog's Qdb; in the future, flows will no longer exist and is_elementary
        will be a trivial function of an exchange asking whether its termination is a context or not.
        :param context:
        :return: bool
        """
        return self._tm[context.fullname].elementary
    '''

    def cascade(self, origin):
        """
        Generate a new query for the specified origin.
        Enables the query to follow the origins of foreign objects found locally.
        If not found locally, the current query is used instead
        :param origin:
        :return:
        """
        return self._grounded_query(origin)

    def _grounded_query(self, origin):
        if origin is None or origin == self._origin:
            return self
        return self._catalog.query(origin, cache=False)

    '''
    def __str__(self):
        return '%s for %s (catalog: %s)' % (self.__class__.__name__, self.origin, self._catalog.root)
    '''
    def _setup_background(self, bi):
        self._debug('Setting up background interface')
        try:
            bi.setup_bm(self)
        except AttributeError:
            raise BackgroundSetup('Failed to configure background')

    def _iface(self, itype, strict=False):
        if self._catalog is None:
            raise NoCatalog
        returned = set()
        if itype in self._iface_cache:
            self._debug('Returning cached iface')
            i = self._iface_cache[itype]
            yield i
            returned.add(i._archive)
        for i in self._catalog.gen_interfaces(self._origin, itype, strict=strict):
            if i._archive in returned:
                continue
            if itype == 'background':  # all our background implementations must provide setup_bm(query)
                self._setup_background(i)

            self._debug('yielding %s' % i)
            if itype not in self._iface_cache:  # only cache the first iface
                if i.origin == self.origin:  # only cache iface if it is a strict match
                    self._iface_cache[itype] = i
            yield i

    def _perform_query(self, itype, attrname, exc, *args, strict=False, **kwargs):
        if itype is None:
            raise BadInterfaceSpec(itype, attrname)  # itype = 'basic'  # fetch, get properties, uuid, reference

        self._debug('Performing %s query, origin %s, iface %s' % (attrname, self.origin, itype))
        message = 'itype %s required for attribute %s' % (itype, attrname)
        props = []
        run = 0
        try:
            for iface in self._iface(itype, strict=strict):
                run += 1
                try:
                    self._debug('Attempting %s query on iface %s' % (attrname, iface))
                    result = getattr(iface, attrname)(*args, **kwargs)
                    message = '(%s) %s' % (itype, attrname)  # implementation found
                except exc:  
                    message = '(%s) %s except %s' % (itype, attrname, exc.__name__)
                    continue
                except NotImplementedError:  # allow nonimplementations to pass silently
                    message = '(%s) %s not implemented' % (itype, attrname)
                    continue
                if result is None:  # successful query must return something
                    message = '(%s) %s null' % (itype, attrname)
                else:
                    if attrname == 'properties':
                        for k in result:
                            if k not in props:
                                props.append(k)
                    else:
                        return result
            if attrname == 'get_context':
                if run:
                    return NullContext

        except AttributeError as e:
            message = '(%s) Attribute error %s' % (attrname, e.args)
        if len(props) > 0:
            return props

        raise exc('%s: %s | %s' % (self.origin, message, args))

    def resolve(self, itype=INTERFACE_TYPES, strict=False):
        """
        Secure access to all known resources but do not answer any query
        :param itype: default: all interfaces
        :param strict: [False]
        :return:
        """
        for k in self._iface(itype, strict=strict):
            yield k

    def get(self, eid, **kwargs):
        """
        Retrieve entity by external Id. This will take any interface and should keep trying until it finds a match.
        It first matches canonical entities, because that is the point of canonical entities.
        :param eid: an external Id
        :return:
        """
        try:
            entity = self._perform_query('basic', 'get', EntityNotFound, eid, **kwargs)
            return self.make_ref(entity)
        except EntityNotFound:
            return self._tm.get_canonical(eid)

    def get_reference(self, external_ref):
        ref = self._perform_query('basic', 'get_reference', EntityNotFound, external_ref)
        # quantity: unit
        # flow: quantity
        # process: list
        # fragment: fragment
        # [context: context]
        if ref is None:
            deref = None
        elif isinstance(ref, list):
            deref = [RxRef(self.make_ref(x.process), self.make_ref(x.flow), x.direction, x.comment, x.value)
                     for x in ref]
        elif isinstance(ref, str):
            deref = ref
        elif ref.entity_type == 'unit':
            deref = ref.unitstring
        else:
            deref = self.make_ref(ref)
        return deref

    '''
    LCIA Support
    get_canonical(quantity)
    catch get_canonical calls to return the query from the local Qdb; fetch if absent and load its characterizations
    (using super ==> _perform_query)
    '''
    '''
    def get_context(self, term, **kwargs):
        cx = super(CatalogQuery, self).get_context(term, **kwargs)
        return self._tm[cx]
    '''

    def get_canonical(self, quantity, **kwargs):
        try:
            # print('Gone canonical')
            q_can = self._tm.get_canonical(quantity)
        except EntityNotFound:
            if hasattr(quantity, 'entity_type') and quantity.entity_type == 'quantity':
                print('Missing canonical quantity-- adding to LciaDb')
                self._catalog.register_entity_ref(quantity)
                return self._tm.get_canonical(quantity)
                # print('Retrieving canonical %s' % q_can)
            else:
                raise
        return q_can

    def synonyms(self, item, **kwargs):
        """
        Potentially controversial? include canonical as well as provincial synonyms for catalog queries??
        :param item:
        :param kwargs:
        :return:
        """
        rtn_set = set()
        for i in self._tm.synonyms(item):
            if i not in rtn_set:
                rtn_set.add(i)
                yield i
        for i in super(CatalogQuery, self).synonyms(item, **kwargs):
            if i not in rtn_set:
                rtn_set.add(i)
                yield i

    def characterize(self, flowable, ref_quantity, query_quantity, value, context=None, location='GLO', **kwargs):
        """
        This is an Xdb innovation: we do not need or want an implementation-specific characterize routine-- just like
        with make_ref, the point of the catalog query is to localize all characterizations to the LciaEngine.

        We simply duplicate the characterize() code from the core QuantityImplementation
        :param flowable:
        :param ref_quantity:
        :param query_quantity:
        :param value:
        :param context:
        :param location:
        :param kwargs:
        :return:
        """
        rq = self.get_canonical(ref_quantity)
        qq = self.get_canonical(query_quantity)
        origin = kwargs.pop('origin', self.origin)
        # print('@@@ going characterization-commando')
        return self._tm.add_characterization(flowable, rq, qq, value, context=context, location=location,
                                             origin=origin, **kwargs)

    def clear_seen_characterizations(self, quantity):
        """
        An ugly hack to deal with the absolutely terrible way we are working around our slow-ass Qdb implementation
        the proper solution is for qdb lookup to be local,  fast and correct, so as to not require caching at all.
        :param quantity:
        :return:
        """
        for i in self._iface_cache.values():
            if i._archive:
                for f in i._archive.entities_by_type('flow'):
                    k = [cf for cf in f._chars_seen.keys() if cf[0] is quantity]
                    for cf in k:
                        f.pop_char(*cf)
                    if f._query_ref:
                        k = [cf for cf in f._query_ref._chars_seen.keys() if cf[0] is quantity]
                        for cf in k:
                            f._query_ref.pop_char(*cf)

    def make_ref(self, entity):
        if isinstance(entity, list):
            return [self.make_ref(k) for k in entity]
        if entity is None:
            return None
        if entity.is_entity:
            try:
                e_ref = entity.make_ref(self._grounded_query(entity.origin))
            except UnknownOrigin:
                e_ref = entity.make_ref(self)
        else:
            e_ref = entity  # already a ref
        if entity.entity_type == 'quantity':

            '''# question of whether to put this test before or after the get_canonical() attempts
            if before- we allow non-canonical "mass" to accumulate in the workspace
            if after- we catch 'mass' and 'freight' (and any LCIA methods we already know locally)
            and only return shit we don't recognize, but lose connection to remote sources
            
            prefer before: we may want to query core quantities like "net calorific value" for flow characterizations
            '''
            if entity.has_lcia_engine():  # we don't need a canonical version if it runs itself
                return e_ref

            ''' # astonishingly, we don't want this - register but not return
            # print('Going canonical')
            # astonishing because it's not true. 
            Well. not exactly true.
            
            CatalogQueries should return canonical quantities. that is the point of the catalog.  The reason we didn't
            want this was because we were using the catalog to access origin-specific data to re-serve it.  On the
            server side, we thought we would want to keep track of all this- for veracity of the data, for provenance,
            etc.  But in point of fact, there is NO CIRCUMSTANCE under which a user benefits from having 
            origin-specific  versions of "mass" or "area".
            
            True, the data won't match the source.  but we will still RECOGNIZE the source because we will register the 
            quantity terms with the term manager.  Which we WEREN"T doing before.
            
            2024-08-10 we are officially abandoning the following corollary: just send back canonical
            A corollary of this is that CatalogQuery.get() should get_canonical FIRST
            '''
            try:
                return self._tm.get_canonical(entity)  # returns the existing canonical unless there is a unit conflict
            except EntityNotFound:
                self._catalog.register_entity_ref(e_ref)
                return self._tm.get_canonical(entity)
        else:
            return e_ref

    '''
    de-reference Characterization factor models
    '''

    def _resolve_cf(self, cf: CharacterizationModel) -> Characterization:
        """

        :param cf:
        :return:
        """
        rq = self.get_canonical(cf.ref_quantity)
        qq = self.get_canonical(cf.query_quantity)
        try:
            cx = self._tm[tuple(cf.context)]
        except InconsistentLineage:
            cx = NullContext
        c = Characterization(cf.flowable, rq, qq, cx, origin=cf.origin)
        for k, v in cf.value.items():
            c[k] = v
        return c

    def factors(self, quantity, flowable=None, context=None, **kwargs):
        for cf in super(CatalogQuery, self).factors(quantity, flowable=flowable, context=context, **kwargs):
            if isinstance(cf, CharacterizationModel):
                yield self._resolve_cf(cf)
            else:
                yield cf

    '''
    de-reference LciaResult models
    '''

    def _result_from_model(self, process_ref, quantity, res_m: LciaResultModel):
        """
        Constructs a Detailed LCIA result from a background LCIA query, when we don't have a list of exchanges
        :param process_ref:
        :param quantity:
        :param res_m:
        :return:
        """
        try:
            quantity = self._tm.get_canonical(res_m.quantity.external_ref)
        except EntityNotFound:
            quantity = CatalogRef.from_json(res_m.quantity.serialize())  # dumbest thing ever- (the de/re/deserialize)

        res = LciaResult(quantity, scenario=res_m.scenario, scale=res_m.scale)
        process = self.get(process_ref)
        for c in res_m.components:
            try:
                entity = self.cascade(c.origin).get(c.entity_id)
            except (UnknownOrigin, EntityNotFound):
                entity = c.component
            res.add_component(c.component, entity)
            for d in c.details:
                try:
                    value = d.result / d.factor.value
                except ZeroDivisionError:
                    value = 0.0
                cx = self._tm[tuple(d.factor.context)]
                try:
                    rq = self.get_canonical(d.exchange.quantity_ref)
                except EntityNotFound:
                    rq = self.get(d.exchange.quantity_ref)
                try:
                    # OK - here we are making redundant orphan FlowRefs, one per detail, to avoid excess API calls.
                    flow_ref = FlowRef(d.exchange.external_ref, self.cascade(d.exchange.origin),
                                       name=d.exchange.name, reference_entity=rq, context=cx)
                except UnknownOrigin:
                    flow_ref = FlowRef(d.exchange.external_ref, self, masquerade=d.exchange.origin,
                                       name=d.exchange.name, reference_entity=rq, context=cx)
                ex_dir = d.exchange.direction
                '''
                if ex_dir is None:
                    print('Bad context: %s' % list(cx))
                    print('using "Output" direction')
                    ex_dir = 'Output'
                '''
                if cx.sense:
                    if ex_dir != comp_dir(cx.sense):
                        value *= -1  # DetailedLciaResult will negate the result internally
                ex = ExchangeRef(process, flow_ref,
                                 ex_dir,
                                 termination=cx, value=value)
                cf = QRResult(d.factor.flowable, rq, quantity, cx,
                              d.factor.locale, d.factor.origin, d.factor.value)
                if ex.flow.quell_co2:
                    co2_cf = CO2QuantityConversion(cf)
                    res.add_score(c.component, ex, co2_cf)
                else:
                    res.add_score(c.component, ex, cf)
        for s in res_m.summaries:
            try:
                entity = self.cascade(s.origin).get(s.entity_id)
            except (UnknownOrigin, EntityNotFound):
                entity = s.component
            res.add_summary(s.component, entity, s.node_weight, s.unit_score)
        if len(res) == 0:
            res.add_summary('Total', 'total', 1.0, res_m.total)
        else:
            if res.total() != res_m.total:
                net = res_m.total - res.total()
                try:
                    res.add_summary('net result', 'net result', 1.0, net)
                except MixedComponents:
                    logging.warning('Inconsistent Agg LCIA result (net %g of %g)' % (net, res_m.total))
        return res

    def _cycle_through_ress(self, ress, process, query_qty):
        if isinstance(ress, list):
            conv = []
            for res in ress:
                if isinstance(res, LciaResultModel):
                    conv.append(self._result_from_model(process_ref=process, quantity=query_qty, res_m=res))
                else:
                    conv.append(res)
            return conv
        else:
            if isinstance(ress, LciaResultModel):
                return self._result_from_model(process_ref=process, quantity=query_qty, res_m=ress)
            else:
                return ress

    def contrib_lcia(self, process, quantity=None, ref_flow=None, **kwargs):
        p = self.get(process)
        r = p.reference(ref_flow)
        ffs = FragmentFlow.from_process_inventory(self, p, r.flow, **kwargs)
        res = frag_flow_lcia(ffs, quantity)
        return res.contrib()

    def sys_lcia(self, process, query_qty, observed=None, ref_flow=None, **kwargs):
        """
        Reimplement this to detect pydantic LciaResult models and de-reference them
        :param process:
        :param query_qty:
        :param observed:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        ress = super(CatalogQuery, self).sys_lcia(process, query_qty, observed=observed, ref_flow=ref_flow, **kwargs)
        return self._cycle_through_ress(ress, process, query_qty)

    def bg_lcia(self, process, query_qty=None, ref_flow=None, **kwargs):
        """
        returns an LciaResult object, aggregated as appropriate depending on the interface's privacy level.
        This can only be implemented at the query level because it requires access to lci()
        :param process: must have a background interface
        :param query_qty: an operable quantity_ref, or catalog default may be used if omitted
        :param ref_flow:
        :param kwargs:
        :return:
        """
        try:
            ress = super(CatalogQuery, self).bg_lcia(process, query_qty=query_qty, ref_flow=ref_flow, **kwargs)
        except BasicRequired:
            p_ref = self.get(process)
            if p_ref.is_entity:
                raise NotImplementedError  # we can't proceed
            lci = p_ref.lci(ref_flow=ref_flow)
            # aggregation
            ress = query_qty.do_lcia(lci, **kwargs)
        return self._cycle_through_ress(ress, process, query_qty)



from ..entities.fragments import Fragment


class OldTraverser(Fragment):
    def exchange_value(self):
        pass

    def traverse(self, scenario=None, observed=False, frags_seen=None):
        if isinstance(scenario, set):
            scenarios = set(scenario)
        elif isinstance(scenario, tuple) or isinstance(scenario, list):
            scenarios = set(scenario)
        elif scenario is None:
            if observed:
                scenarios = {1}
            else:
                scenarios = None
        else:
            scenarios = {scenario}
        ffs, _ = self._traverse_node(1.0, scenarios, frags_seen=frags_seen)
        return ffs

    def _traverse_fg_node(self, ff, scenarios, frags_seen):
        """
        Handle foreground nodes and processes--> these can be quantity-conserving, but except for
        balancing flows the flow magnitudes are determined at the time of construction (or scenario specification).

        Balancing flow exchange values are always determined with respect to a unit activity of the terminal node.

        This is messy so it deserves some notes.
        the fragment's exchange value specifies the scaling factor for the terminal node, EXCEPT if the
        fragment is a reference fragment (no parent) AND the terminal node is in the foreground.  In this case
        ONLY, the exchange value functions as an inbound exchange value, and the node weight should be 1.0.
        The prototypical example of this is a fragment for cultivation of 1 ha of corn, with the reference
        flow being (e.g.) 9300 kg of corn.  The reference fragment's exchange value is 9300 kg; the reference
         node's node weight is 1.0

        If the reference node has a balancing flow (say, we want to balance carbon content), then the stock
        value is the reference flow converted to the conservation quantity, e.g. 9300 * 0.43 = 3999 kg C, sign
        negative since it's going out (note that the reference fragment's direction is 'Input' in this case
        because direction is always interpreted relative to the parent).

        So then we traverse the child flows, let's say none of them have kg C characterization, and so our stock
        remains -3999 kg. WHen we hit the balancing fragment "atmospheric carbon in", we catch the FoundBalanceFlow
        and come back to it.

        When we come back, we re-negate the stock to +3999 and pass that as _balance to the balancing flow, which
        becomnes that flow's exchange value (again w.r.t. this node's unit node weight).

        IF the terminal node is a process, or if the node is an interior (non-reference) fragment, it's much easier.
        The stock is simply the process's inbound exchange value (with respect to a unit activity level), or if
        it's a foreground node then the stock is simply 1, and the node_weight already accounts for the exchange
        value and scales the balancing flow correctly.

        Were we to do this non-'live' or with some kind of stack, we could conceivably balance multiple quantities with
        additional fragment flows, but I'm pretty sure the same effect can be achieved by nesting conserving fragments.
        This also requires the modeler to specify a resolution order and cuts the need for elaborate error checking.

        :param ff: a FragmentFlow containing the foreground termination to recurse into
        :param scenarios:
        :param frags_seen:
        :return: a list of FragmentFlows in the order encountered, with input ff in position 0
        """
        term = ff.term
        node_weight = ff.node_weight
        ffs = [ff]
        stock_inflow = stock_outflow = 0.0  # keep track of total inflows to detect+quash near-zero (floating-point-tolerance) balances

        if term.is_fg or not term.valid:
            if self.reference_entity is None:
                # inbound exchange value w.r.t. term node's unit magnitude
                stock = self.exchange_value(scenarios)
            else:
                stock = 1.0  # balance measurement w.r.t. term node's unit magnitude
        else:
            stock = term.inbound_exchange_value  # balance measurement w.r.t. term node's unit magnitude
        bal_f = None
        if self.is_conserved_parent:
            stock *= self.balance_magnitude
            if self.direction == 'Input':  # convention: inputs to self are positive
                #  direction w.r.t. parent so self.direction == 'Input' is an input to parent = output from node
                stock *= -1
            self.dbg_print('%g inbound-balance' % stock, level=2)

        if stock > 0:
            stock_inflow += stock
        else:
            stock_outflow -= stock
        # TODO: Handle auto-consumption!  test child flows for flow identity and null term; cumulate; scale node_weight

        for f in self.child_flows:
            try:  # now that we are storing _balance_child we can just skip it instead of try-catch. but whatever.
                # traverse child, collecting conserved value if applicable
                child_ff, cons = f._traverse_node(node_weight, scenarios,
                                                  frags_seen=frags_seen, conserved_qty=self.conserved_quantity)
                if cons is None:
                    self.dbg_print('-- returned cons_value', level=3)
                else:
                    self.dbg_print('%g returned cons_value' % cons, level=2)
                    stock += cons
                    if cons > 0:
                        stock_inflow += cons
                    else:
                        stock_outflow -= cons
            except FoundBalanceFlow:
                self.dbg_print('%g bal magnitude on %.3s' % (stock, f.uuid), level=3)
                bal_f = f
                child_ff = []

            ffs.extend(child_ff)

        stock_max = max([abs(stock_inflow), abs(stock_outflow)])

        if bal_f is not None:
            # balance reports net inflows; positive value is more coming in than out
            # if balance flow is an input, its exchange must be the negative of the balance
            # if it is an output, its exchange must equal the balance
            if stock != 0 and (abs(stock) / stock_max < 1e-10):
                self.dbg_print('Quashing <1e-10 balance flow (%g vs %g)' % (stock, stock_inflow), level=1)
                stock = 0.0
            if bal_f.direction == 'Input':
                stock *= -1
                self.dbg_print('%.3s Input: negating balance value' % bal_f.uuid)
            else:
                self.dbg_print('%.3s Output: maintaining balance value' % bal_f.uuid)
            self.dbg_print('%g balance value passed to %.3s' % (stock, bal_f.uuid))
            bal_ff, _ = bal_f._traverse_node(node_weight, scenarios,
                                             frags_seen=set(frags_seen), conserved_qty=None, _balance=stock)
            ffs.extend(bal_ff)

        return ffs

    def _traverse_subfragment(self, ff, scenarios, frags_seen):
        """
        handle sub-fragments, including background flows--
        for sub-fragments, the flow magnitudes are determined at the time of traversal and must be pushed out to
         child flows
        for LOCAL background flows, the background ff should replace the current ff, maintaining self as fragment

        subfragment activity level is determined as follows:
         - if the subfragment is a background fragment, it MUST have a unity inbound_exchange_value; this is enforced:
           - for background processes, because FlowTermination._unobserved_exchanges() uses term_node.lci(ref_flow)
           - for background fragments, in the yet-to-be-implemented bg_lcia method  (part of the foreground interface??)
           in any case, the background term is swapped into the foreground node.

         - otherwise, the subfragment inventory is taken and grouped, and the matching flow is found, and the magnitude
           of the matching flow is used to normalize the downstream node weight.

         - then the inventory of the subfragment is used to apply exchange values to child fragments, and traversal
           recursion continues.


        :param ff: a FragmentFlow containing the non-fg subfragment termination to recurse into
        :param scenarios:
        :param frags_seen:
        :return:
        """
        '''
        '''

        term = ff.term
        if 0:  #  term.term_is_bg: this is deprecated
            pass
            '''
            if term.term_flow == term.term_node.flow:
                # collapse trivial bg terminations into the parent fragment flow
                bg_ff, _ = term.term_node._traverse_node(ff.node_weight, scenario, observed=observed)
                assert len(bg_ff) == 1, (self.uuid, term.term_node.external_ref)
                bg_ff[0].fragment = self
                return bg_ff
            '''

        # traverse the subfragment, match the driven flow, compute downstream node weight and normalized inventory
        try:
            ffs, unit_inv, downstream_nw = _do_subfragment_traversal(ff, scenarios, frags_seen)
        except ZeroInboundExchange:
            self.dbg_print('subfragment divide by zero', 1)
            return [ff]

        # next we traverse our own child flows, determining the exchange values from the normalized unit inventory
        for f in self.child_flows:
            self.dbg_print('Handling child flow %s' % f, 4)
            ev = 0.0
            try:
                m = next(j for j in unit_inv if j.fragment.flow == f.flow)
                if m.fragment.direction == f.direction:
                    self.dbg_print('  ev += %g' % m.magnitude, 4)
                    ev += m.magnitude
                else:
                    self.dbg_print('  ev -= %g' % m.magnitude, 4)
                    ev -= m.magnitude
                unit_inv.remove(m)
            except StopIteration:
                self.dbg_print('  no driving flows found')
                continue

            self.dbg_print('traversing with ev = %g' % ev, 4)
            child_ff, _ = f._traverse_node(downstream_nw, scenarios,
                                           frags_seen=frags_seen, _balance=ev)
            ffs.extend(child_ff)

        # remaining un-accounted io flows are getting appended, so do scale
        for x in unit_inv:
            x.scale(downstream_nw)
        ffs.extend(list(unit_inv))

        return ffs

    def _traverse_node(self, upstream_nw, scenarios,
                       frags_seen=None, conserved_qty=None, _balance=None):

        """
        If the node has a non-null termination, use that; follow child flows.

        If the node's termination is null- then look for matching background fragments. If one is found, adopt its
        termination, and return.

        else: assume it is a null foreground node; follow child flows

        :param upstream_nw: upstream node weight
        :param scenarios: set of scenario values
        #:param observed: whether to use observed or cached evs (overridden by scenario specification)
        :param frags_seen: carried along to catch recursion loops
        :param conserved_qty: in case the parent node is a conservation node
        :param _balance: used when flow magnitude is determined during traversal, i.e. for balance flows and
        children of fragment nodes
        :return: 2-tuple: ffs, conserved_val
          ffs = an array of FragmentFlow records reporting the traversal, beginning with self
          conserved_val = the magnitude of the flow with respect to the conserved quantity, if applicable (or None)
        """

        # first check for cycles
        if frags_seen is None:
            frags_seen = set()

        if self.reference_entity is None:
            if self.uuid in frags_seen:
                # this should really get resolved into a loop-closing algorithm
                raise InvalidParentChild('Frag %s seeing self\n %s' % (self.uuid, '; '.join(frags_seen)))
            frags_seen.add(self.uuid)

        if _balance is None:
            _scen_ev = self._match_scenario_ev(scenarios)
            ev = self.exchange_value(_scen_ev)
        else:
            _scen_ev = None
            self.dbg_print('%g balance' % _balance, level=2)
            ev = _balance
            '''
            if self._check_observability(scenario):
                self._cache_balance_ev(_balance, scenario, observed)
            '''

        magnitude = upstream_nw * ev
        _scen_term = self._match_scenario_term(scenarios)
        term = self._terminations[_scen_term]

        if self.reference_entity is None:
            node_weight = upstream_nw
            if magnitude == 0:
                self.dbg_print('Unobserved reference fragment')
                node_weight = 0
        else:
            node_weight = magnitude

        f_conv = term.node_weight_multiplier

        node_weight *= f_conv

        self.dbg_print('magnitude: %g f_conv: %g node_weight: %g' % (magnitude, f_conv, node_weight))

        conserved_val = None
        if conserved_qty is not None:
            if self.is_balance:
                self.dbg_print('Found balance flow')
                raise FoundBalanceFlow  # to be caught
            cf = conserved_qty.cf(self.flow)
            self.dbg_print('consrv cf %g for qty %s' % (cf, conserved_qty), level=3)
            conserved_val = ev * cf
            if conserved_val == 0:
                conserved = False
            else:
                conserved = True
                if self.direction == 'Output':  # convention: inputs to parent are positive
                    conserved_val *= -1
                self.dbg_print('conserved_val %g' % conserved_val, level=2)
        elif self.is_balance:
            # traversing balance flow after FoundBalanceFlow exception
            conserved = True
        else:
            conserved = self.conserved

        # print('%6f %6f %s' % (magnitude, node_weight, self))
        # this is the only place a FragmentFlow is created
        # TODO: figure out how to cache + propagate matched scenarios ... in progress
        ff = FragmentFlow(self, magnitude, node_weight, term, conserved, match_ev=_scen_ev, match_term=_scen_term,
                          flow_conversion=f_conv)

        '''
        now looking forward: is our termination a cutoff, background, foreground or subfragment?
        '''
        if magnitude == 0:
            # no flow to follow
            self.dbg_print('zero magnitude')
            return [ff], conserved_val
        elif term.is_null or term.is_context:
            # cutoff /context and background end traversal
            self.dbg_print('cutoff or bg')
            return [ff], conserved_val

        if term.is_fg or term.is_process or not term.valid:
            self.dbg_print('fg')
            ffs = self._traverse_fg_node(ff, scenarios, frags_seen)

        else:
            self.dbg_print('subfrag')
            ffs = self._traverse_subfragment(ff, scenarios, frags_seen)

        return ffs, conserved_val


def _do_subfragment_traversal(ff, scenarios, frags_seen):
    """
    This turns out to be surprisingly complicated. So we now have:
     - LcFragment._traverse_node <-- which is called recursively
      + scenario matching + finds ev and term
      - selects handler based on term type:
       - LcFragment._subfragment_traversal
       |- invokes (static) _do_subfragment_traversal (YOU ARE HERE)
       ||- calls [internally recursive] term_node.unit_flows, which is just a wrapper for
       || - (static) group_ios
       ||  + reference flow and autoconsumption handling
       || /
       ||/
       |+ reference flow matching and normalizing
       + child flow matching and further recursion --> into LcFragment._traverse_node
      /
     /
     ffs, conserved_val


     - (static) group_ios
     - nested inside LcFragment.unit_flows, which is really just a wrapper
     - called from
     - called from

    :param ff: The FragmentFlow whose subfragment is being traversed
    :param scenarios: to pass to subfragment
    :return: ffs [subfragment traversal appended], unit_inv [with reference flow removed], downstream node weight
    """
    term = ff.term  # note that we ignore term.direction in subfragment traversal
    node_weight = ff.node_weight
    self = ff.fragment

    unit_inv, subfrags = term.term_node.unit_flows(scenario=scenarios, frags_seen=frags_seen)

    # find the inventory flow that matches us
    # use term_flow over term_node.flow because that allows client code to specify inverse traversal knowing
    #  only the sought flow.
    # unit_flows guarantees that there is exactly one of these flows (except in the case of cumulating flows!
    # see group_ios)
    try:
        match = next(k for k in unit_inv if k.fragment.flow == term.term_flow)
    except StopIteration:
        print('Flow mismatch Traversing:\n%s' % self)
        print('Term flow: %s' % term.term_flow.link)
        print(term.serialize())
        for k in unit_inv:
            print('%s' % k.fragment.flow.link)
        raise MissingFlow('Term flow: %s' % term.term_flow.link)
    self.dbg_print('matched flow %s' % match.fragment.flow.link)

    unit_inv.remove(match)

    in_ex = match.magnitude  # this is the inbound exchange value for the driven fragment
    if in_ex == 0:
        # this indicates a non-consumptive pass-thru fragment.
        print('Frag %.5s: Zero inbound exchange' % self.uuid)
        raise ZeroInboundExchange
    if match.fragment.direction == self.direction:  # match direction is w.r.t. subfragment
        # self is driving subfragment in reverse
        self.dbg_print('reverse-driven subfragment %.3s' % match.fragment.uuid)
        in_ex *= -1

    # node weight for the driven [downstream] fragment
    downstream_nw = node_weight / in_ex

    '''
    # then we add the results of the subfragment, either in aggregated or disaggregated form
    if term.descend:
        # if appending, we are traversing in situ, so do scale
        self.dbg_print('descending', level=0)
        for i in subfrags:
            i.scale(downstream_nw)
        ffs = [ff]
        ffs.extend(subfrags)
    else:
        # if aggregating, we are only setting unit scores- so don't scale
        self.dbg_print('aggregating', level=0)
        ff.aggregate_subfragments(subfrags, scenarios=scenarios)  # include params to reproduce
        ff.node_weight = downstream_nw  # NOTE: this may be problematic; undone in lca_disclosures
        ffs = [ff]
    '''
    # now, we abolish descend as a traversal parameter and make it only an LCIA parameter
    # therefore, we retain subfragments always, we just have to decide how to scale them
    self.dbg_print('%d subfrags resulting from this traversal' % len(subfrags), 1)
    ff.aggregate_subfragments(subfrags, scenarios=scenarios)
    ff.node_weight = downstream_nw
    ffs = [ff]

    return ffs, unit_inv, downstream_nw
